import numpy as np
import pandas as pd
from tqdm import tqdm
from joblib import Parallel, delayed

def cytopath_merger(adata, overlap=0.5, num_cores=1):
    """
    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix with end points.
    overlap: float (default: 0.5)
        overlap until new bucket of cells is created.
    Returns
    -------
    adata.uns['trajectories']["cells_along_trajectories_each_step_merged"]: List of arrays containing the cell indexes and allignment scores for each step

    """
    cells_along_trajectories = adata.uns['trajectories']["cells_along_trajectories_each_step"].copy()
    cells_along_trajectories_merged = []
    
    # For each terminal region and each trajectory
    for end_point in adata.uns['run_info']['end_point_clusters']:
        print('Merging steps for trajectories for ' + end_point)
        sel_end_point_data = cells_along_trajectories[np.where(cells_along_trajectories["End point"]==end_point)[0]]
        for i in tqdm(range(adata.uns['run_info']["trajectory_count"][end_point])):
            # Find all trajectories
            sel_end_point_trajectories=sel_end_point_data[np.where(sel_end_point_data["Trajectory"]==i)[0]]
            start=0
            steps=0
            global_merge_steps=[]
            # Create list, which steps to merge
            
            # Retrieve all steps with % overlap
            while steps < len(np.unique(sel_end_point_trajectories["Step"])):
                merge_steps = [start]
                steps += 1
                intersect_step = len(np.intersect1d(sel_end_point_trajectories[np.where(sel_end_point_trajectories["Step"]==start)[0]]["Cell"],sel_end_point_trajectories[np.where(sel_end_point_trajectories["Step"]==steps)[0]]["Cell"]))
                total_unqiue = len(np.unique(np.append(sel_end_point_trajectories[np.where(sel_end_point_trajectories["Step"]==start)[0]]["Cell"],sel_end_point_trajectories[np.where(sel_end_point_trajectories["Step"]==steps)[0]]["Cell"])))
                overlap_perc = intersect_step/(total_unqiue+1e-15)
                if overlap_perc > overlap:
                        merge_steps.append(steps)
                while overlap_perc > overlap and steps < len(np.unique(sel_end_point_trajectories["Step"])):
                    steps += 1
                    intersect_step = len(np.intersect1d(sel_end_point_trajectories[np.where(sel_end_point_trajectories["Step"]==start)[0]]["Cell"],sel_end_point_trajectories[np.where(sel_end_point_trajectories["Step"]==steps)[0]]["Cell"]))
                    total_unqiue = len(np.unique(np.append(sel_end_point_trajectories[np.where(sel_end_point_trajectories["Step"]==start)[0]]["Cell"],sel_end_point_trajectories[np.where(sel_end_point_trajectories["Step"]==steps)[0]]["Cell"])))
                    overlap_perc = intersect_step/total_unqiue
                    if overlap_perc > overlap:
                        merge_steps.append(steps)
                    
                start = steps
                global_merge_steps.append(merge_steps)              

            counter=0
            # Merge all adjacent steps
            for h in range(len(global_merge_steps)):
                for k in range(len(global_merge_steps[h])):
                    end_p = cells_along_trajectories["End point"] == end_point
                    average_t = cells_along_trajectories["Trajectory"] == i
                    step_av = cells_along_trajectories["Step"] == k + counter
                    end_p_avg = np.zeros((len(end_p)))

                    for l in range(len(end_p_avg)):
                        end_p_avg[l] = end_p[l] and average_t[l] and step_av[l]
                        cells_along_trajectories["Step"][np.where(end_p_avg)[0]] = h                   
                counter+=len(global_merge_steps[h])

            # Save new bags of cell under separate key
            for m in range(len(global_merge_steps)):
                end_p = cells_along_trajectories["End point"]==end_point
                average_t = cells_along_trajectories["Trajectory"]==i
                step_av = cells_along_trajectories["Step"]==m
                end_p_avg = np.zeros((len(end_p)))

                for l in range(len(end_p_avg)):
                    end_p_avg[l] = end_p[l] and average_t[l] and step_av[l]

                cells = np.unique(cells_along_trajectories["Cell"][np.where(end_p_avg)[0]])
                cell_along_traj = cells_along_trajectories[np.where(end_p_avg)[0]]

                for n in range(len(cells)):
                    max_a = np.max(cell_along_traj[np.where(cell_along_traj["Cell"]==cells[n])]["Allignment Score"])
                    cells_along_trajectories_merged.append([end_point, i, m, cells[n], max_a])
                    
    adata.uns['trajectories']["cells_along_trajectories_each_step_merged"] = np.rec.fromrecords(cells_along_trajectories_merged, 
                                                                                                dtype=[('End point', 'U8'), 
                                                                                                       ('Trajectory', int), 
                                                                                                       ('Step', int), 
                                                                                                       ('Cell', 'U8'), 
                                                                                                       ('Allignment Score', float)])




from . import anndata
from . import scvelo
from .markov_chain_sampler import sampling
from .trajectory_inference import trajectories, merge_steps
from .plotting import plot_trajectories, plot_sampling
from .utils import read_processed


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.animation as ani
      
def frame_gen(i, adata, basis, samples):
    plt.scatter(adata.obsm['X_'+basis][np.unique(samples[:, i-1]),0], adata.obsm['X_'+basis][np.unique(samples[:, i-1]),1], c='green', s=100)

def pl_cytopath_sampling(adata, basis="umap", folder=""):

    if folder[-1] != '/':
        folder+='/'

    fig = plt.figure()
    plt.scatter(adata.obsm['X_'+basis][:,0], adata.obsm['X_'+basis][:,1], c='grey', s=100)
    samples = adata.uns['samples']['cell_sequences']
    animator = ani.FuncAnimation(fig, frame_gen, frames= samples.shape[1], fargs=(adata, basis, samples), interval = 200)
    animator.save(folder+'sampling.gif')

    


import numpy as np
import pandas as pd
from tqdm import tqdm
from scipy import spatial
from scvelo.preprocessing.neighbors import get_connectivities
from sklearn.metrics.pairwise import cosine_similarity
from joblib import Parallel, delayed

def surrogate_cell_neighborhood_finder(adata, end_point, map_state, fill_cluster=True, n_neighbors=30, cluster_freq=0.1, mode='distances', recurse_neighbors=True):
    """Finds surrogate cell and uses the scvelo neighborhood graph to finds its recursive neighbors. Applicable when mean is used to compute trajectory coordinates. Default is median.
    
    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix with end points.
    end_point: `string`
        End point cluster to which the trajectories belong
    map_state: matrix
        Coordinates of cells in selected basis.
    n_neighbors: `integer`(default: 30)
        Number of neighbors to consider

    Returns
    -------
    Returns list of arrays containing the sequence of neighborhoods along the trajectory and the sequence of surrogate cells.
    """
    # Locate surrogate cells to anchor trajectory
    final_cluster=adata.uns['trajectories']["trajectories_coordinates"][end_point]
    final_sequences=[]
    for i in tqdm(range(len(final_cluster))):
        final_cluster_sequence=final_cluster['trajectory_'+str(i)+'_coordinates']
        cell_sequences=[]
        for j in range(len(final_cluster_sequence)):
            cell_sequences.append(spatial.KDTree(map_state).query(final_cluster_sequence[j])[1])
        final_sequences.append(cell_sequences)

    # Neighborhood of surrogate cells
    neighborhood=get_connectivities(adata, mode=mode, n_neighbors=n_neighbors, recurse_neighbors=recurse_neighbors)
    neighborhood_sequence=[]
    for i in tqdm(range(len(final_sequences))):
        local_neighs=[]
        for j in range(len(final_sequences[i])):
            local_neighs.append(neighborhood[final_sequences[i][j]].nonzero()[1])
            
            if fill_cluster == True:
                print('Fill cluster not implemented.')

        neighborhood_sequence.append(local_neighs)
    return neighborhood_sequence, cell_sequences

def cell_neighborhood_finder(adata, map_state, end_point, neighbors_basis='pca', fill_cluster=True, n_neighbors_cluster=30, cluster_freq = 0.5, n_neighbors='auto'):
    """Finds the nearest neighbors along the average trajectory.
    
    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix with end points.
    end_point: `string`
        End point cluster to which the trajectories belong
    fill_cluster: Boolean (default:True)
        Enforce only cells in compostional clusters are assigned score.
    n_neighbors_cluster: `integer` (Default:30)
        Number of cells to consider for determining compositional clusters
    cluster_freq: `float` (Default: 0.25)
        Frequency of cell cluster cells per step to consider as compositonal cluster
    n_neighbors: `` (default: 'auto')
        Number of cells to consider for alignment scoring.

    Returns
    -------
    Returns list of arrays containing the sequence of neighborhoods along the trajectory.
    """

    # Cell neighborhood of trajectory
    final_cluster=adata.uns['trajectories']["trajectories_coordinates"][end_point]
    
    # Find clusters composing the trajectory
    compositional_clusters = []
    for i in tqdm(range(len(final_cluster))):
        final_cluster_sequence=final_cluster['trajectory_'+str(i)+'_coordinates']
        for j in range(len(final_cluster_sequence)):
            cell_sequences_ = spatial.KDTree(map_state).query(final_cluster_sequence[j], k=n_neighbors_cluster)[1]
            compositional_clusters_ = adata.obs.loc[adata.obs.index.values[cell_sequences_], adata.uns['run_info']['cluster_key']].astype(str)
            compositional_clusters_ = compositional_clusters_.value_counts()
            compositional_clusters_ /= sum(compositional_clusters_)
            compositional_clusters_ = compositional_clusters_.loc[compositional_clusters_ >= cluster_freq]
            compositional_clusters.extend(compositional_clusters_.index.tolist())
    adata.uns['trajectories']['compositional_clusters'][end_point] = np.unique(compositional_clusters)

    # Find neighbors in PCA space
    if neighbors_basis=='pca':

        # Use pca to find neigbhors if it exists
        map_state_ = adata.obsm['X_pca']

        final_cluster_ = {}
        for i in range(len(final_cluster)):
            final_cluster_sequence=final_cluster['trajectory_'+str(i)+'_coordinates']
            final_cluster_sequence_ = []
            for j in range(len(final_cluster_sequence)):
                final_cluster_sequence_.append(adata.obsm['X_pca'][spatial.KDTree(map_state).query(final_cluster_sequence[j])[1]])
            final_cluster_['trajectory_'+str(i)+'_coordinates'] = final_cluster_sequence_
        final_cluster = final_cluster_
    else:
        map_state_ = map_state

    
    # Neighborhood size
    if n_neighbors == 'auto':
        n_neighbors = adata.obs.loc[adata.obs[adata.uns['run_info']['cluster_key']].astype(str).isin(adata.uns['trajectories']['compositional_clusters'][end_point]), adata.uns['run_info']['cluster_key']].value_counts().max()

    # Create neighborhood
    neighborhood_sequence=[]
    for i in tqdm(range(len(final_cluster))):
        final_cluster_sequence=final_cluster['trajectory_'+str(i)+'_coordinates']
        cell_sequences=[]
        for j in range(len(final_cluster_sequence)):
            if fill_cluster == False:
                cell_sequences.append(spatial.KDTree(map_state_).query(final_cluster_sequence[j], k=n_neighbors)[1])
            elif fill_cluster == True:
                cell_sequences_ = spatial.KDTree(map_state_).query(final_cluster_sequence[j], k=n_neighbors)[1]
                cell_sequences.append(cell_sequences_[np.where(adata.obs.loc[adata.obs.index.values[cell_sequences_], adata.uns['run_info']['cluster_key']].astype(str).isin(adata.uns['trajectories']['compositional_clusters'][end_point]))[0]])
        neighborhood_sequence.append(cell_sequences)
    return neighborhood_sequence
    
def directionality_score(adata, neighborhood_sequence, end_point, map_state, num_cores=1): 

    """Scores the directionality of the cells along the path:
    Evaluation using the cosine angle between trajectory and neighboring cell to cell transitions and
    the transition probability
    
    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix with end points.
    end_point: `string`
        End point cluster to which the trajectories belong
    map_state: matrix
        Coordinates of cells in selected basis.
    neighborhood_sequence: 
        List of the cells along the trajectory.
        
    Returns
    -------
    Returns list of arrays containing the scores of all cells at each step.
    """

    final_cluster=adata.uns['trajectories']["trajectories_coordinates"][end_point]
    
    transmat=adata.uns["T_forward"].copy()
    transmat[transmat.nonzero()]=1
    
    if 'velocity_graph' and 'velocity_graph_neg' in adata.uns.keys():
        velocity_graph=adata.uns["velocity_graph"]+adata.uns["velocity_graph_neg"]
        
    velocity_graph=velocity_graph.multiply(transmat)
    all_scores=[]
    def parallelize(j):
        score=np.zeros((len(neighborhood_sequence[i][j])))

        if j==0:
            avg_difference=final_cluster['trajectory_'+str(i)+'_coordinates'][j+1]-final_cluster['trajectory_'+str(i)+'_coordinates'][j]
            neighborhood_transitions=velocity_graph[neighborhood_sequence[i][j]]

            for k in range(len(neighborhood_sequence[i][j])):
                cell_jumps=neighborhood_transitions[k,:].nonzero()[1]
                origin_cell=map_state[neighborhood_sequence[i][j][k]]
                end_cells=map_state[cell_jumps]
                difference=origin_cell-end_cells

                # Score based on alignment and distance 
                score[k]=np.average(cosine_similarity(avg_difference.reshape(1,-1),difference)*np.exp(neighborhood_transitions[k,:].data))

        elif j!=(len(neighborhood_sequence[i])-1) and j!=0: 
            avg_difference_forward=final_cluster['trajectory_'+str(i)+'_coordinates'][j+1]-final_cluster['trajectory_'+str(i)+'_coordinates'][j]
            avg_difference_backward=final_cluster['trajectory_'+str(i)+'_coordinates'][j]-final_cluster['trajectory_'+str(i)+'_coordinates'][j-1]

            neighborhood_transitions=velocity_graph[neighborhood_sequence[i][j]]
            for k in range(len(neighborhood_sequence[i][j])):
                cell_jumps=neighborhood_transitions[k,:].nonzero()[1]
                origin_cell=map_state[neighborhood_sequence[i][j][k]]
                end_cells=map_state[cell_jumps]
                difference=origin_cell-end_cells

                # Score based on alignment and distance 
                score_forward=np.average(cosine_similarity(avg_difference_forward.reshape(1,-1), difference)*np.exp(neighborhood_transitions[k,:].data))
                score_backward=np.average(cosine_similarity(avg_difference_backward.reshape(1,-1), difference)*np.exp(neighborhood_transitions[k,:].data))
                
                score[k]=max(score_forward, score_backward)

        elif j==(len(neighborhood_sequence[i])-1):
            avg_difference_backward=final_cluster['trajectory_'+str(i)+'_coordinates'][j]-final_cluster['trajectory_'+str(i)+'_coordinates'][j-1]

            neighborhood_transitions=velocity_graph[neighborhood_sequence[i][j]]
            for k in range(len(neighborhood_sequence[i][j])):
                cell_jumps=neighborhood_transitions[k,:].nonzero()[1]
                origin_cell=map_state[neighborhood_sequence[i][j][k]]
                end_cells=map_state[cell_jumps]
                difference=origin_cell-end_cells

                # Score based on alignment and distance  
                score_backward=np.average(cosine_similarity(avg_difference_backward.reshape(1,-1), difference)*np.exp(neighborhood_transitions[k,:].data))
                score[k]=score_backward

        return (j, score)

    # Calculation of the directionality score for each cell in each trajectory
    for i in range(len(final_cluster)):
        cluster_scores = Parallel(n_jobs=num_cores)(delayed(parallelize)(j) for j in tqdm(range(0, len(neighborhood_sequence[i]))))
        cluster_scores = dict((index, score) for index, score in cluster_scores)
        all_scores.append([cluster_scores[j] for j in range(0, len(neighborhood_sequence[i]))])

    return all_scores

def cutoff_score(adata, end_point, neighborhood_sequence, all_scores, cut_off=0.0):
    """Deletes all cells with a score lower than the cutoff.
    
    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix with end points.
    map_state: matrix 
        Which projection to use for the data.
    end_point: `integer`
        End point cluster to which the trajectories belong.
    neighborhood_sequence: 
        List of the cells along the trajectory.
    all_scores:
        List of arrays with the score for each cell allong the trajectory
    """
    final_cluster = adata.uns['trajectories']["trajectories_coordinates"][end_point]
    directional_neighborhood_sequence = neighborhood_sequence
    if cut_off != None:
        for i in tqdm(range(len(final_cluster))):
            cutoff_score = cut_off
            for j in range(0, len(neighborhood_sequence[i])):
                 directional_neighborhood_sequence[i][j] = np.delete(directional_neighborhood_sequence[i][j], np.where(all_scores[i][j]<=cutoff_score))
                 all_scores[i][j] = np.delete(all_scores[i][j], np.where(all_scores[i][j]<=cutoff_score))
    return directional_neighborhood_sequence, all_scores
    
def cytopath(adata, basis="umap", neighbors_basis='pca', surrogate_cell=False, fill_cluster=True, n_neighbors_cluster=30, cluster_freq=0.25, n_neighbors='auto', cut_off=0.0, num_cores=1):
    """Calculates the average time step for each cell.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix with end points.
    basis: str/list (default: umap)
        The space in which the neighboring cells should be searched. If None imputed expression is used. If list, imputed expression from supplied genes is used.
    surrogate_cell: Boolean (default:False)
        Whether or not a surrogate cell should be used for the neighborhood search
    n_neighbors:  str/int/float (default:'auto')
        Number of neighbors to searched along the average trajectory.
    cut_off_percentile: float (default:0.0)
        Cuttof for the directionality score along the average trajectory for the cells.
    Returns
    -------
    Returns adata.uns['trajectories']["cells_along_trajectories"]: List of arrays, which denote the average step for cell.
    adata.uns['trajectories']["cells_along_trajectories_each_step"]: List of arrays containing the cell indexes for each step
    """
    cell_score=[]
    step_ordering_trajectory=[]
    cells_trajectory=[]
    
    adata.uns['trajectories']['compositional_clusters'] = {}
    
    if basis == None:
        map_state = adata.layers['Ms']
    elif type(basis) == list:
        map_state = adata[:, basis].layers['Ms']
    else:
        map_state = adata.obsm['X_'+basis]

    for end_point_cluster in adata.uns['trajectories']["trajectories_coordinates"].keys():
        if surrogate_cell==True:
            print('Anchoring trajectories for end point  ' + end_point_cluster +' to cells in dataset and computing neighborhoods')
            neighborhood_sequence, cell_sequences = surrogate_cell_neighborhood_finder(adata, end_point_cluster, map_state, mode='distances', fill_cluster=fill_cluster, n_neighbors_cluster=n_neighbors_cluster,
                                                                                    cluster_freq=cluster_freq, n_neighbors=n_neighbors, recurse_neighbors=True)
        else:
            print('Computing neighborhoods of trajectories for end point ' + end_point_cluster +' at each step')
            neighborhood_sequence = cell_neighborhood_finder(adata, map_state, end_point=end_point_cluster, neighbors_basis=neighbors_basis, fill_cluster=fill_cluster, n_neighbors_cluster=n_neighbors_cluster, cluster_freq=cluster_freq,
                                                             n_neighbors=n_neighbors)

        print('Computing alignment score of cells in trajectory neighborhood w.r.t. trajectories for end point  ' + end_point_cluster)
        all_scores=directionality_score(adata, neighborhood_sequence=neighborhood_sequence, end_point=end_point_cluster, map_state=map_state, num_cores=num_cores)

        print('Removing cells below cutoff threshold from trajectories for end point  ' + end_point_cluster +' (i.e. cells neighborhood)')
        directional_neighborhood_sequence, all_scores = cutoff_score(adata, end_point=end_point_cluster, neighborhood_sequence=neighborhood_sequence, 
                                                                  all_scores=all_scores, cut_off=cut_off)
        cell_score.append(all_scores)       
        cells_trajectory.append(directional_neighborhood_sequence)

    cells_along_trajectories = []
    end_point_clusters = list(adata.uns['trajectories']["trajectories_coordinates"].keys())
    for i in range(len(cells_trajectory)):
        for j in range(len(cells_trajectory[i])):
            for k in range(len(cells_trajectory[i][j])):
                for l in range(len(cells_trajectory[i][j][k])):
                    cells_along_trajectories.append((end_point_clusters[i], j, k, cells_trajectory[i][j][k][l], cell_score[i][j][k][l]))

    adata.uns['trajectories']["cells_along_trajectories_each_step"] = np.rec.fromrecords(cells_along_trajectories, 
                                                                                         dtype=[('End point', 'U32'), ('Trajectory', int), 
                                                                                                ('Step', int), ('Cell', int), ('Allignment Score', float)])
def estimate_cell_data(adata, groupby='median', weighted=False):
    """Calculates the pseudotime per trajectory for each cell.
    
    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix with end points. Must run cytopath.cytopath before.
    groupby:  str(default: max)
        One of max, min, median or mean. Grouping method for determing alignment score per cell per trajectory
    weighted: Boolean (default:False)
        If groupby is mean, reports mean pseduotime weighted by alignment score if true.
    """
    
    if 'cells_along_trajectories_each_step' not in adata.uns['trajectories'].keys():
        raise ValueError("Run cytopath.cytopath before estimating pseudotime.")
        
    cytopath_data = pd.DataFrame(adata.uns['trajectories']['cells_along_trajectories_each_step'].copy())
    
    if groupby == 'max':
        cytopath_data_ = cytopath_data.loc[cytopath_data['Allignment Score']==cytopath_data.groupby(['End point', 'Trajectory', 'Cell'])["Allignment Score"].transform('max')]
        cytopath_data_ = cytopath_data_.groupby(['End point', 'Trajectory', 'Cell']).mean()
        
    elif groupby == 'min':
        cytopath_data_ = cytopath_data.loc[cytopath_data['Allignment Score']==cytopath_data.groupby(['End point', 'Trajectory', 'Cell'])["Allignment Score"].transform('min')]
        cytopath_data_ = cytopath_data_.groupby(['End point', 'Trajectory', 'Cell']).mean()
       
    elif groupby == 'median':
        cytopath_data_ = cytopath_data.loc[cytopath_data['Allignment Score']==cytopath_data.groupby(['End point', 'Trajectory', 'Cell'])["Allignment Score"].transform('median')]
        cytopath_data_ = cytopath_data_.groupby(['End point', 'Trajectory', 'Cell']).mean()
    
    elif groupby == 'mean' and weighted==False:
        cytopath_data_ = cytopath_data
        cytopath_data_ = cytopath_data_.groupby(['End point', 'Trajectory', 'Cell']).mean()

    elif groupby == 'mean' and weighted==True:
        cytopath_data_ = cytopath_data
        cytopath_data_['Step'] = (cytopath_data_["Step"]*cytopath_data_['Allignment Score'])/cytopath_data_.groupby(['End point', 'Trajectory', 'Cell'])["Allignment Score"].transform('sum')
        step_ = cytopath_data_.groupby(['End point', 'Trajectory', 'Cell'])['Step'].sum()

        cytopath_data_ = cytopath_data_.groupby(['End point', 'Trajectory', 'Cell']).mean()
        cytopath_data_['Step'] = step_

    adata.uns['trajectories']['cells_along_trajectories'] = cytopath_data_.to_records()
    
    adata.uns['trajectories']['step_time'] = pd.DataFrame(index=np.arange(adata.shape[0])).merge(cytopath_data_.reset_index().pivot(index='Cell', columns=['End point', 'Trajectory'], 
                                                                                                                                    values='Step'), 
                                                                                                 left_index=True, right_index=True, how='left').T.values
    
    
    cell_fate_prob = pd.DataFrame(index=np.arange(adata.shape[0])).merge(cytopath_data_.reset_index().pivot(index='Cell', columns=['End point', 'Trajectory'], values='Allignment Score'), 
                                                                         left_index=True, right_index=True, how='left')
    cell_fate_prob = cell_fate_prob.fillna(0).div(cell_fate_prob.sum(axis=1), axis=0)    
    adata.uns['trajectories']['cell_fate_probability'] = cell_fate_prob.T.values
    
    
    

    




import numpy as np
from scipy import spatial
import matplotlib.pyplot as plt
      
def pl_cytopath_alignment(adata, basis="umap", folder="", figsize=(12,3), save_type='svg', size = 10, smoothing=False):
    
    map_state = adata.obsm['X_'+basis]
    av_allign_score_glob=[]
    std_allign_score_glob=[]
    
    step_time = adata.uns['trajectories']['step_time']
    fate_prob = adata.uns['trajectories']['cell_fate_probability']

    sequence=0
   
    for end_point_cluster in adata.uns['run_info']["end_point_clusters"]:
        trajectories = adata.uns['trajectories']["cells_along_trajectories_each_step"][np.where(adata.uns['trajectories']["cells_along_trajectories_each_step"]["End point"]==end_point_cluster)[0]]
        for i in range(adata.uns['run_info']['trajectory_count'][end_point_cluster]):

            av_trajectories=trajectories[np.where(trajectories["Trajectory"]==i)[0]]
            av_allign_score=np.zeros((len(np.unique(av_trajectories["Step"]))))
            std_allign_score=np.zeros((len(np.unique(av_trajectories["Step"]))))
            for l in range(len(np.unique(av_trajectories["Step"]))):
                av_allign_score[l]=np.average((av_trajectories[np.where(av_trajectories["Step"]==l)[0]]["Allignment Score"]))
                std_allign_score[l]=np.std((av_trajectories[np.where(av_trajectories["Step"]==l)[0]]["Allignment Score"]))
            
            # TODO: Separate per step average alignment score calculation from plotting

            # Plotting
            path = folder+"_end_point_"+end_point_cluster+"_cytopath_"+str(i)+\
                 "occurance"+str(adata.uns['run_info']["trajectories_sample_counts"][end_point_cluster][i])+"."+save_type

            fig, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3, figsize=figsize) 
            ax1.plot(range(len(np.unique(av_trajectories["Step"]))), av_allign_score, color='black')
            ax1.fill_between(range(len(np.unique(av_trajectories["Step"]))), 
                             av_allign_score+std_allign_score, av_allign_score-std_allign_score, facecolor='grey', alpha=0.6)
            ax1.set_ylabel('Mean/std. of alignment scores per step')
            ax1.set_xlabel('Steps')
            
            # Plot step size for aligned cells
            sc_step = ax2.scatter(map_state[:,0], map_state[:,1], alpha=0.6, s=size, color="grey")
            sc_step = ax2.scatter(map_state[:,0], map_state[:,1], alpha=0.9, s=size, c=step_time[sequence,:], cmap='YlGnBu')
            fig.colorbar(sc_step, ax=ax2, label='Step time')

            ax2.set_ylabel(basis.upper()+' 2')
            ax2.set_xlabel(basis.upper()+' 1')

            # Plot alignment score
            sc_score = ax3.scatter(map_state[:,0], map_state[:,1], alpha=0.6, s=size, color="grey")
            sc_score = ax3.scatter(map_state[:,0], map_state[:,1], alpha=0.9, s=size, c=fate_prob[sequence,:], cmap='YlGnBu')
            fig.colorbar(sc_score, ax=ax3, label='Cell fate probability')

            ax3.set_ylabel(basis.upper()+' 2')
            ax3.set_xlabel(basis.upper()+' 1')

            # Plot trajectory
            if basis in adata.uns['run_info']['trajectory_basis']:

                coords = np.array(adata.uns['trajectories']['trajectories_coordinates'][end_point_cluster]['trajectory_'+str(i)+'_coordinates'])

                if smoothing == True:
                    #TODO: More relevant procedure required
                    signal = coords[:,0] + 1j*coords[:,1]

                    # FFT and frequencies
                    fft = np.fft.fft(signal)
                    freq = np.fft.fftfreq(signal.shape[-1])

                    # filter
                    cutoff = 0.1
                    fft[np.abs(freq) > cutoff] = 0

                    # IFFT
                    signal_filt = np.fft.ifft(fft)

                    coords[:,0] = signal_filt.real
                    coords[:,1] = signal_filt.imag

                ax2.plot(coords[:, 0], coords[:, 1], color='black')
                ax3.plot(coords[:, 0], coords[:, 1], color='black')

            elif ('pca' in adata.uns['run_info']['trajectory_basis']) and (basis != 'pca'):

                coords_ = np.array(adata.uns['trajectories']['trajectories_coordinates'][end_point_cluster]['trajectory_'+str(i)+'_coordinates'])

                cell_sequences=[]
                for j in range(len(coords_)):
                    cell_sequences.append(spatial.KDTree(adata.obsm['X_pca']).query(coords_[j])[1])
                
                coords = map_state[cell_sequences]

                if smoothing == True:
                    #TODO: More relevant procedure required
                    signal = coords[:,0] + 1j*coords[:,1]

                    # FFT and frequencies
                    fft = np.fft.fft(signal)
                    freq = np.fft.fftfreq(signal.shape[-1])

                    # filter
                    cutoff = 0.1
                    fft[np.abs(freq) > cutoff] = 0

                    # IFFT
                    signal_filt = np.fft.ifft(fft)

                    coords[:,0] = signal_filt.real
                    coords[:,1] = signal_filt.imag

                ax2.plot(coords[:, 0], coords[:, 1], color='black')
                ax3.plot(coords[:, 0], coords[:, 1], color='black')

            fig.savefig(path, bbox_inches='tight', dpi=300)
            # End plotting

            sequence+=1

            av_allign_score_glob.append(av_allign_score)                                
            std_allign_score_glob.append(std_allign_score)


from cytopath import anndata

def read_processed(data_loc):

    adata = anndata.read(data_loc)

    try:
        if type(adata.uns['run_info']['end_point_clusters']) == str:
            adata.uns['run_info']['end_point_clusters'] = [adata.uns['run_info']['end_point_clusters']]
    except:
        pass
    
    try:
        for key in adata.uns['run_info']['trajectories_sample_counts'].keys():
            if type(adata.uns['run_info']['trajectories_sample_counts'][key]) != list:
                adata.uns['run_info']['trajectories_sample_counts'][key] = [adata.uns['run_info']['trajectories_sample_counts'][key]]
    except:
        pass

    try:
        adata.uns['trajectories']['cells_along_trajectories_each_step'] = \
        adata.uns['trajectories']['cells_along_trajectories_each_step'].astype([('End point', 'U8'), ('Trajectory', int), ('Step', int), ('Cell', int), 
                                                                                ('Allignment Score', float)])
    except:
        pass

    try:
        adata.uns['trajectories']['cells_along_trajectories_each_step_merged'] = \
        adata.uns['trajectories']['cells_along_trajectories_each_step_merged'].astype([('End point', 'U8'), ('Trajectory', int), ('Step', int), ('Cell', int), 
                                                                                ('Allignment Score', float)])
    except:
        pass

    return adata


import math
import numpy as np
from scipy import sparse
from tqdm import tqdm
from joblib import Parallel, delayed
import warnings

def matrix_prep(trans_matrix):

    # For increased speed: Create and save np array of the cumsum on nonzero entries-->80% faster
    trans_matrix_ = sparse.csr_matrix(trans_matrix)
    trans_matrix_indices = np.split(trans_matrix_.indices, trans_matrix_.indptr)[1:-1]
    trans_matrix_probabilites = np.split(trans_matrix_.data, trans_matrix_.indptr)[1:-1]
    
    for i in range(len(trans_matrix_probabilites)):
        trans_matrix_probabilites[i]=np.cumsum(trans_matrix_probabilites[i])
    
    # For increased speed: Save data and indices of nonzero entries in np array--> 210% faster
    trans_matrix_indice_zeros = np.zeros((len(trans_matrix_probabilites), max(np.count_nonzero(trans_matrix, axis=1))), dtype=np.int32)
    trans_matrix_probabilites_zeros = np.zeros((len(trans_matrix_probabilites), max(np.count_nonzero(trans_matrix, axis=1))), dtype=np.float32)
    
    for i in range(len(trans_matrix_probabilites)):
        for j in range(len(trans_matrix_probabilites[i])):
            trans_matrix_indice_zeros[i, j] = trans_matrix_indices[i][j]
            trans_matrix_probabilites_zeros[i, j] = trans_matrix_probabilites[i][j]
    
    # FIXME: Ensure that cummulative sum reaches 1 precisely (rounding to 3 decimals ensures that this almost never happens)
    return trans_matrix_indice_zeros, np.round(trans_matrix_probabilites_zeros, decimals=3)

# Markov sim function: sample according to the probabilities in the transition matrix.
# Gets used in the parallel function.
def markov_sim(j, sim_number, max_steps, root_cells, clusters, trans_matrix, trans_matrix_indice_zeros, trans_matrix_probabilites_zeros):
    # Create empty arrays for the simulation and sample random numbers for all samples and max steps
    prob_state = np.empty((sim_number, max_steps-1), dtype=np.float32)
    currState = np.empty((sim_number, max_steps), dtype=np.int32)
    clust_state = np.empty((sim_number, max_steps), dtype='U8')

    currState[:, 0] = root_cells[j]
    clust_state[:, 0] = clusters[root_cells[j]]
    random_step = np.random.rand(sim_number, max_steps-1)
    
    for i in range(1, max_steps):
        for k in range(sim_number):
            # Retrieve next cell transition and save the probabilities of the transition and the cluster the cell is in
            currState[k, i] = trans_matrix_indice_zeros[currState[k, i-1], 
                                                        np.where(random_step[k, i-1] <= trans_matrix_probabilites_zeros[currState[k,i-1], :])[0][0]]
            prob_state[k, i-1] = trans_matrix[currState[k, i-1], currState[k, i]]
            clust_state[k, i] = clusters[currState[k, i]]
    return currState, np.sum(-np.log(prob_state), axis=1), clust_state

def sampling(data, matrix_key = 'T_forward', cluster_key = 'louvain', max_steps=200, traj_number=1000, sim_number=2000,
                end_point_probability=0.95, root_cell_probability=0.95, end_points=[], root_cells=[], end_clusters = [], root_clusters=[], min_clusters=3,
                normalize=False, unique=True, num_cores=1, copy=False):
    
    """Markov sampling of cell sequences starting from defined root cells to defined terminal regions based on a cell-cell transition probability matrix.

    Arguments
    ---------
    adata: :class:`~anndata.AnnData`
        Annotated data matrix with transition probabilities, end points and clustering.
    matrix_key: Key for accessing transition matrix stored in adata.uns
    cluster_key: Clustering of cells to use for terminal region annotation.  
    max_steps: `integer` (default: 200)
        Maximum number steps of simulation.
    traj_number: `integer`(default: 1000)
        Minimum number of samples generated for each terminal region.
    sim_number: `integer`(default: 2000)
        Initial number of samples generated for each terminal region.
    end_point_probability: 'float' (default:0.95)
        End point probability threshold at which a cell is considered to be a terminal state for samples.
    root_state_probability: 'float' (default:0.95)
        Start state probability threshold at which a cell is considered to be an origin for samples.
    end_points: `list` (default: [])
        Numerical indices of the cells considered to be end points for manual annotation. Precendence over end_point_probability and end_clusters.
    root_cells: `list` (default: [])
        Numerical indices of the cells considered to be root states for manual annotation. Precendence over root_state_probability and root_clusters.
    end_clusters: `list` (default: [])
        Cluster IDs to be considered end points. Precendence over end_point_probability.
    root_clusters: `list` (default: [])
        Cluster IDs to be considered root states. Precendence over root_state_probability.
    min_clusters: `integer` (default:3)
        Minium number of clusters covered by each simulation. (cluster_key)
    normalize: 'Boolean' (default: False)
        Toggle row sum normalization.
    unique: 'Boolean' (default:True)
        Only keep unique samples.
    num_cores: 'integer' (default:1)
        Number of cpu cores to use.
    copy: 'Boolean' (default: False)
        Create a copy of the anndata object or work inplace.
    
    Returns
    -------
    adata.uns["samples"]["cell_sequences"]: List of arrays containing the index of the cells in a sequence.
    adata.uns["samples"]["transition_score"]: List with all of the transition probabilities.
    adata.uns["samples"]["cluster_sequences"]: The sequence of clusters to which the cells belong.
    """

    adata = data.copy() if copy else data

    ### Run checks

    # Check if transition matrix exists
    try:
        adata.uns[matrix_key]
    except NameError:
        raise ValueError('Transition matrix must be provided in adata.uns[matrix_key].')
    
    # Check if a square matrix is provided
    try:
        assert len(adata.uns[matrix_key].shape) == 2
        assert adata.uns[matrix_key].shape[0] == adata.uns[matrix_key].shape[1]
        assert adata.uns[matrix_key].shape[0] == adata.shape[0]
    except:
        raise ValueError('Transition matrix must be of shape num_cells x num_cells')

    # Convert sparse matrx to numpy array    
    if sparse.issparse(adata.uns[matrix_key]):
        trans_matrix = np.float64(np.asarray(adata.uns[matrix_key].toarray()))
    else:
        trans_matrix = np.float64(np.asarray(adata.uns[matrix_key]))
    
    if sparse.issparse(adata.X):
        adata.X = np.asarray(adata.X.toarray())
    
    # Check if matrix has been normalized
    if normalize == False:
        row_sums = np.sum(trans_matrix, axis=1)
        try: 
            assert max(row_sums) <= 1.0 + 1e-3
            assert min(row_sums) >= 1.0 - 1e-3
        except AssertionError:
            raise ValueError('Transition matrix has not been normalized. Either pass normalize==True or provide a row sum normalized matrix.')

    # Row sum normalization 
    if normalize == True:
        new_array = trans_matrix
        normalization_v = new_array.sum(axis=1)
        for i in range(len(normalization_v)):
            trans_matrix[i,:]=trans_matrix[i,:]/normalization_v[i]
        
    # Check if cluster labels are provided
    try:
        adata.obs[cluster_key]
        assert adata.obs[cluster_key].dtype == 'category'
    except:
        raise ValueError('Categorical cluster labels must be provided in adata.obs[cluster_key]')
    
    # Use louvain warning
    if cluster_key != 'louvain':
        warnings.warn('A finely resolved clustering is required to identify terminal regions. Consider using louvain clustering.')

    ### Checks Done

    # Define root and end points, clusters
    clusters = adata.obs[cluster_key].astype(str).values
    adata.uns['run_info'] = {'cluster_key': cluster_key}

    trans_matrix_indice_zeros, trans_matrix_probabilites_zeros = matrix_prep(trans_matrix)

    # Retrieve all root cells
    if len(root_cells) == 0:
        if len(root_clusters) == 0:
            try:
                root_cells = np.asarray(np.where(np.asarray(adata.obs["root_cells"]) >= root_cell_probability))[0]
            except:
                raise ValueError('Root cells must be provided in adata.obs["root_cells"]. Run cytopath.terminal_states')
        else:
            root_cells = np.asarray(np.where(np.asarray(adata.obs[cluster_key].isin(root_clusters))))[0]
    adata.uns['run_info']['root_cells'] = root_cells

    # Retrieve all end point cells
    if len(end_points) == 0:
        if len(end_clusters) == 0:
            try:
                end_points = np.asarray(np.where(np.asarray(adata.obs["end_points"]) >= end_point_probability))[0]
            except:
                raise ValueError('End points must be provided in adata.obs["end_points"]. Run cytopath.terminal_states')
        else:
            end_points = np.asarray(np.where(np.asarray(adata.obs[cluster_key].isin(end_clusters))))[0]
    adata.uns['run_info']['end_points'] = end_points
                
    end_point_clusters = adata.obs[cluster_key][end_points].astype(str).values
    end_clusters_ = np.unique(end_point_clusters)
    
    # Initialize all empty lists
    traj_num = [0]
    glob_all_seq = []
    glob_prob_all_seq = []
    glob_cluster_seq_all = []
    sim_number_ = sim_number

    count = 1
    # Resample samples until the number of trajectories for each end point has been reached.
    while min(traj_num) < traj_number:

        print("Sampling round " + str(count))
        count+=1

        # TODO: Add transition probability matrix update that removes transitions exclusive to end points sampled adequately 
              # to increase sampling efficiency for larger datasets.

        # Parallelisation over the cells: 1 Thread for the samples of each cell.
        if len(root_cells) == 1:
            all_seq, prob_all_seq, cluster_seq_all = markov_sim(0, sim_number_, max_steps, root_cells, clusters, trans_matrix,
                                                                trans_matrix_indice_zeros, trans_matrix_probabilites_zeros)
        else:
            results = Parallel(n_jobs=num_cores)(delayed(markov_sim)(j, sim_number_, max_steps, root_cells, clusters, trans_matrix,
                                                                                trans_matrix_indice_zeros, 
                                                                                trans_matrix_probabilites_zeros) for j in tqdm(range(len(root_cells))))

            # Results are stored in np.arrays
            all_seq = np.empty((len(root_cells)*sim_number_, max_steps), dtype=np.int32)
            prob_all_seq = np.empty((len(root_cells)*sim_number_), dtype=np.float32)
            cluster_seq_all = np.empty((len(root_cells)*sim_number_, max_steps), dtype='U32')

            for i in range(len(root_cells)):
                all_seq[(i*sim_number_):((i*sim_number_)+sim_number_), :] = results[i][0]
                prob_all_seq[(i*sim_number_):((i*sim_number_)+sim_number_)] = results[i][1]
                cluster_seq_all[(i*sim_number_):((i*sim_number_)+sim_number_), :] = results[i][2]
            
            del results
               
        # Filter samples that visit less than min_clusters clusters
        indices = [idx for idx in np.arange(all_seq.shape[0]) if np.unique(cluster_seq_all[idx]).shape[0] >= min_clusters]
        if len(indices) == 0:
            raise ValueError('Zero samples obtained. Try increasing max_steps.')
        all_seq = [all_seq[i] for i in sorted(indices)]
        prob_all_seq = [prob_all_seq[i] for i in sorted(indices)]
        cluster_seq_all = [cluster_seq_all[i] for i in sorted(indices)]

        # If unique, then only keep unique samples
        if unique == True:
            indices = np.unique(all_seq, axis=0, return_index=True)[1]
            all_seq = [all_seq[i] for i in sorted(indices)]
            prob_all_seq = [prob_all_seq[i] for i in sorted(indices)]
            cluster_seq_all = [cluster_seq_all[i] for i in sorted(indices)]
        
        glob_all_seq.append(all_seq)
        glob_prob_all_seq.append(prob_all_seq)
        glob_cluster_seq_all.append(cluster_seq_all)
        
        # Count the number of samples to each end point
        all_seq = np.vstack(glob_all_seq)
        traj_num=[]
        for i in range(len(end_clusters_)):
            end_cluster_indices = np.concatenate(np.where(end_point_clusters == end_clusters_[i]), axis=0)
            # Retrieve only sequences, which end in the end points of the cluster
            indices_cluster=[]
            sample_end_points = np.asarray([item[-1] for item in all_seq])
            
            for l in range(len(end_cluster_indices)):
                indices_cluster.append(np.where(sample_end_points == end_points[end_cluster_indices[l]]))

            indices_cluster = np.concatenate(indices_cluster, axis=1)
            all_seq_cluster = all_seq[indices_cluster]
            all_seq_cluster = all_seq_cluster[0]
            traj_num.append(len(all_seq_cluster))
            
        sim_number_ = math.ceil(traj_number/(min(traj_num) + 1))*sim_number
    
    # Retrieve traj_number of samples for each end point.
    glob_prob_all_seq = np.concatenate(glob_prob_all_seq)
    all_seq=np.vstack(glob_all_seq)
    cell_sequences = []
    transition_score = []
    cluster_sequences = []

    for i in range(len(end_clusters_)):
        end_cluster_indices = np.concatenate(np.where(end_point_clusters == end_clusters_[i]), axis=0)
        indices_cluster = []
        sample_end_points = np.asarray([item[-1] for item in all_seq])
        
        for l in range(len(end_cluster_indices)):
            indices_cluster.append(np.where(sample_end_points == end_points[end_cluster_indices[l]]))
        indices_cluster = np.concatenate(indices_cluster, axis=1)
        all_seq_cluster = all_seq[indices_cluster]
        all_seq_cluster = all_seq_cluster[0]
        idx_range = np.random.randint(all_seq_cluster.shape[0], size=traj_number)
        
        cell_sequences.append(all_seq_cluster[idx_range])
        
        loc_tran_score = glob_prob_all_seq[indices_cluster[0]]
        transition_score.append(loc_tran_score[idx_range])

        loc_cluster_seq = np.vstack(glob_cluster_seq_all)[indices_cluster[0]]
        cluster_sequences.append(loc_cluster_seq[idx_range])
    
    # Save results
    adata.uns['run_info']['didrun'] = True

    adata.uns['samples'] = {}
    adata.uns['samples']["cell_sequences"] = np.vstack(cell_sequences)
    adata.uns['samples']["transition_score"] = np.concatenate(transition_score)
    adata.uns['samples']["cluster_sequences"]=np.vstack(cluster_sequences)

    return adata if copy else None


from .trajectory_estimation.sample_clustering import sample_clustering
from .trajectory_estimation.cyto_path import cytopath, estimate_cell_data
from .trajectory_estimation.cytopath_merger import cytopath_merger

def trajectories(data, smoothing=False, alignment_cutoff=0.0, basis='umap', neighbors_basis='pca', fill_cluster=True, n_neighbors_cluster=30, cluster_freq=0.1,
                 groupby='median', weighted=False, surrogate_cell=False, n_neighbors_alignment='auto', cluster_num=1, method = 'kmeans', distance = 'cosine', num_cores=1, copy=False):
    
    adata = data.copy() if copy else data
    
    if 'X_pca' not in adata.obsm.keys(): raise ValueError('Compute PCA using cytopath.scvelo.pp.pca')

    # Clustering of cell sequences to obtain trajectories
    sample_clustering(adata, basis=basis, smoothing=smoothing, 
                          cluster_num=cluster_num, method=method, distance=distance)

    # Align cells along trajectories
    cytopath(adata, basis=basis, neighbors_basis=neighbors_basis, surrogate_cell=surrogate_cell, fill_cluster=fill_cluster, cluster_freq=cluster_freq, n_neighbors_cluster=n_neighbors_cluster,
             n_neighbors=n_neighbors_alignment, cut_off=alignment_cutoff, num_cores=num_cores)
    
    # Estimate pseudotime, cell fate prob and alignment score at cell level
    estimate_cell_data(adata, groupby=groupby, weighted=weighted)

    return adata if copy else None

def merge_steps(data, overlap=0.0, num_cores=1, copy=False):

    adata = data.copy() if copy else data
    cytopath_merger(adata, overlap=overlap, num_cores=num_cores)   
    return adata if copy else None



from setuptools import setup, find_packages

setup(
  name='cytopath',
  version='0.1.6.post2',
  description='Simulation based inference of differentiation trajectories from RNA velocity fields.',
  license='BSD 3-Clause License',
  packages=find_packages(),
  author = 'Revant Gupta',
  author_email = 'revant.gupta.93@gmail.com',
  url = 'https://github.com/aron0093/cytopath',
  download_url = 'https://github.com/aron0093/cytopath/archive/v_016_2.tar.gz',
  keywords = ['Trajectory Inference', 'single-cell RNA sequencing', 'RNA velocity'],
  install_requires=[
          'numpy>=1.20.0',
          'scipy',
          'anndata',
          'scvelo>=0.1.25',
          'joblib',
          'fastdtw',
          'hausdorff',
          'tqdm',
	  'scikit-network'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Science/Research',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ])
